#include "darwin_robot.h"
#include "darwin_robot_exceptions.h"
#include <iostream>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#ifdef _HAVE_XSD
#include "xml/darwin_config.hxx"
#endif

#define       PI       3.14159

const std::string servo_names[MAX_NUM_SERVOS]={std::string("Servo0"),
                                               std::string("j_shoulder_pitch_r"),
                                               std::string("j_shoulder_pitch_l"),
                                               std::string("j_shoulder_roll_r"),
                                               std::string("j_shoulder_roll_l"),
                                               std::string("j_elbow_r"),
                                               std::string("j_elbow_l"),
                                               std::string("j_hip_yaw_r"),
                                               std::string("j_hip_yaw_l"),
                                               std::string("j_hip_roll_r"),
                                               std::string("j_hip_roll_l"),
                                               std::string("j_hip_pitch_r"),
                                               std::string("j_hip_pitch_l"),
                                               std::string("j_knee_r"),
                                               std::string("j_knee_l"),
                                               std::string("j_ankle_pitch_r"),
                                               std::string("j_ankle_pitch_l"),
                                               std::string("j_ankle_roll_r"),
                                               std::string("j_ankle_roll_l"),
                                               std::string("j_pan"),
                                               std::string("j_tilt"),
                                               std::string("Servo21"),
                                               std::string("Servo22"),
                                               std::string("Servo23"),
                                               std::string("Servo24"),
                                               std::string("Servo25"),
                                               std::string("Servo26"),
                                               std::string("Servo27"),
                                               std::string("Servo28"),
                                               std::string("Servo29"),
                                               std::string("Servo30")};

const led_info_t leds[GPIO_NUM_LEDS]={{DARWIN_TX_LED_CNTRL,DARWIN_TX_LED_PERIOD_L,GPIO_VALUE,GPIO_TOGGLE,GPIO_BLINK},
                                      {DARWIN_RX_LED_CNTRL,DARWIN_RX_LED_PERIOD_L,GPIO_VALUE,GPIO_TOGGLE,GPIO_BLINK},
                                      {DARWIN_PLAY_LED_CNTRL,DARWIN_PLAY_LED_PERIOD_L,GPIO_VALUE,GPIO_TOGGLE,GPIO_BLINK},
                                      {DARWIN_EDIT_LED_CNTRL,DARWIN_EDIT_LED_PERIOD_L,GPIO_VALUE,GPIO_TOGGLE,GPIO_BLINK},
                                      {DARWIN_MNG_LED_CNTRL,0x0000,GPIO_VALUE,GPIO_TOGGLE,0x00},
                                      {DARWIN_AUX1_LED_CNTRL,DARWIN_AUX1_LED_COLOR_L,GPIO_VALUE_R,GPIO_TOGGLE_R,0x00},
                                      {DARWIN_AUX1_LED_CNTRL,DARWIN_AUX1_LED_COLOR_L,GPIO_VALUE_G,GPIO_TOGGLE_G,0x00},
                                      {DARWIN_AUX1_LED_CNTRL,DARWIN_AUX1_LED_COLOR_L,GPIO_VALUE_B,GPIO_TOGGLE_B,0x00},
                                      {DARWIN_AUX2_LED_CNTRL,DARWIN_AUX2_LED_COLOR_L,GPIO_VALUE_R,GPIO_TOGGLE_R,0x00},
                                      {DARWIN_AUX2_LED_CNTRL,DARWIN_AUX2_LED_COLOR_L,GPIO_VALUE_G,GPIO_TOGGLE_G,0x00},
                                      {DARWIN_AUX2_LED_CNTRL,DARWIN_AUX2_LED_COLOR_L,GPIO_VALUE_B,GPIO_TOGGLE_B,0x00}};


CDarwinRobot::CDarwinRobot(const std::string &name,std::string &bus_id,int bus_speed, unsigned char id)
{
  int num_buses,i;

  this->robot_device=NULL;
  if(name!="")
  {
    this->robot_name=name;
    this->dyn_server=CDynamixelServerFTDI::instance();
    num_buses=this->dyn_server->get_num_buses();
    if(num_buses>0)
    {
      this->dyn_server->config_bus(bus_id,bus_speed);
      this->robot_device=dyn_server->get_device(id,dyn_version2);
      this->robot_id=id;
      /* wait for the firmware to finish the scan */
      do{
        /* get the current manager status */
        this->robot_device->read_byte_register(DARWIN_MM_CNTRL,&this->manager_status);
        usleep(100000);
      }while(this->manager_status&MANAGER_SCANNING);
      /* get the number of servos detected */
      this->robot_device->read_byte_register(DARWIN_MM_NUM_SERVOS,&this->manager_num_servos);
      /* get the current action status */
      this->robot_device->read_byte_register(DARWIN_ACTION_CNTRL,&this->action_status);
      /* get the current smart charger status (detected or not)*/
      this->robot_device->read_byte_register(DARWIN_SMART_CHARGER_CNTRL,&this->smart_charger_status);
      this->MIN_limit_current=0.255;
      this->MAX_limit_current=1.0;
      
    }
    else
    {
      /* handle exceptions */
      throw CDarwinRobotException(_HERE_,"No FTDI USB devices found");
    }
  }
  else
  {
    /* handle exceptions */    
    throw CDarwinRobotException(_HERE_,"Invalid robot name");
  }
}

// GPIO interface
void CDarwinRobot::gpio_set_led(led_t led)
{
  if(this->robot_device!=NULL)
    this->robot_device->write_byte_register(leds[led].control_reg,leds[led].value_mask);
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::gpio_clear_led(led_t led)
{
  if(this->robot_device!=NULL)
    this->robot_device->write_byte_register(leds[led].control_reg,0x00);
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::gpio_toggle_led(led_t led)
{
  if(this->robot_device!=NULL)
    this->robot_device->write_byte_register(leds[led].control_reg,leds[led].toggle_mask);
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::gpio_blink_led(led_t led,int period_ms)
{
  unsigned char data[3];

  if(this->robot_device!=NULL)
  {
    if(led>=LED_TX && led<=LED_3)
    {
      data[0]=leds[led].blink_mask;// blink led
      data[1]=period_ms%256;
      data[2]=period_ms/256;
      this->robot_device->write_registers(leds[led].control_reg,data,3);
    }
    else
      throw CDarwinRobotException(_HERE_,"The desired LED can not be configured to blink");
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::gpio_aux1_color(double red, double green, double blue)
{
  unsigned char data[3];
  unsigned short int color;

  if(this->robot_device!=NULL)
  {
    data[0]=GPIO_COLOR;
    /* saturate the color */
    if(red>1.0) red=1.0;
    else if(red<0.0) red=0.0;
    if(green>1.0) green=1.0;
    else if(green<0.0) green=0.0;
    if(blue>1.0) blue=1.0;
    else if(blue<0.0) blue=0.0;
    color=((unsigned short int)(red*31.0))+(((unsigned short int)(green*31.0))<<5)+(((unsigned short int)(blue*31.0))<<10);
    data[1]=color%256;
    data[2]=color/256;
    this->robot_device->write_registers(DARWIN_AUX1_LED_CNTRL,data,3);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::gpio_aux2_color(double red, double green, double blue)
{
  unsigned char data[3];
  unsigned short int color;

  if(this->robot_device!=NULL)
  {
    data[0]=GPIO_COLOR;
    /* saturate the color */
    if(red>1.0) red=1.0;
    else if(red<0.0) red=0.0;
    if(green>1.0) green=1.0;
    else if(green<0.0) green=0.0;
    if(blue>1.0) blue=1.0;
    else if(blue<0.0) blue=0.0;
    color=((unsigned short int)(red*31.0))+(((unsigned short int)(green*31.0))<<5)+(((unsigned short int)(blue*31.0))<<10);
    data[1]=color%256;
    data[2]=color/256;
    this->robot_device->write_registers(DARWIN_AUX2_LED_CNTRL,data,3);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

bool CDarwinRobot::is_button_pressed(pushbutton_t pushbutton)
{
  unsigned char value;

  if(this->robot_device!=NULL)
  {
    switch(pushbutton)
    {
      case START_PB: this->robot_device->read_byte_register(DARWIN_START_PB_CNTRL,&value);
                     break;
      case MODE_PB: this->robot_device->read_byte_register(DARWIN_MODE_PB_CNTRL,&value);
                    break;
    }
    if(value&GPIO_VALUE)
      return true;
    else
      return false;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

// ADC interface
void CDarwinRobot::adc_start(void)
{
  if(this->robot_device!=NULL)
    this->robot_device->write_byte_register(DARWIN_ADC_CNTRL,ADC_START);
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

bool CDarwinRobot::adc_is_running(void)
{
  unsigned char status;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_byte_register(DARWIN_ADC_CNTRL,&status);
    if(status&ADC_RUNNING)
      return true;
    else
      return false;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::adc_set_period(unsigned char period_ms)
{
  if(this->robot_device!=NULL)
    this->robot_device->write_byte_register(DARWIN_ADC_PERIOD,period_ms);
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}
double CDarwinRobot::adc_get_value(adc_t adc)
{
  unsigned short int value;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_word_register(DARWIN_ADC_CH1_L+((int)adc)*2,&value);

    return ((double)value)/((double)(1<<12));
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

double CDarwinRobot::adc_get_temperature(void)
{
  unsigned short int value;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_word_register(DARWIN_ADC_TEMP_L,&value);

    return ((double)value)/((double)(1<<10));
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

double CDarwinRobot::adc_get_vrefint(void)
{
  unsigned short int value;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_word_register(DARWIN_ADC_VREF_L,&value);

    return ((double)value)/((double)(1<<12));
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::adc_stop(void)
{
  if(this->robot_device!=NULL)
    this->robot_device->write_byte_register(DARWIN_ADC_CNTRL,ADC_STOP);
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

// imu interface
void CDarwinRobot::imu_start(void)
{
  if(this->robot_device!=NULL)
    this->robot_device->write_byte_register(DARWIN_IMU_CNTRL,IMU_START);
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::imu_set_cal_samples(unsigned short int num_samples)
{
  if(this->robot_device!=NULL)
    this->robot_device->write_word_register(DARWIN_IMU_CAL_SAMPLES_L,num_samples);
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

unsigned short int CDarwinRobot::imu_get_cal_samples(void)
{
  unsigned short int samples;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_word_register(DARWIN_IMU_CAL_SAMPLES_L,&samples);
    return samples;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::imu_stop(void)
{
  if(this->robot_device!=NULL)
    this->robot_device->write_byte_register(DARWIN_IMU_CNTRL,IMU_STOP);
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::imu_start_gyro_cal(void)
{
  if(this->robot_device!=NULL)
    this->robot_device->write_byte_register(DARWIN_IMU_CNTRL,IMU_START_CAL);
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

bool CDarwinRobot::imu_is_gyro_cal_done(void)
{
  unsigned char status;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_byte_register(DARWIN_IMU_CNTRL,&status);
    if(status&IMU_CALIBRATING)
      return true;
    else
      return false;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

bool CDarwinRobot::imu_is_accel_detected(void)
{
  unsigned char status;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_byte_register(DARWIN_IMU_CNTRL,&status);
    if(status&IMU_ACCEL_DET)
      return true;
    else 
      return false;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

bool CDarwinRobot::imu_is_gyro_detected(void)
{
  unsigned char status;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_byte_register(DARWIN_IMU_CNTRL,&status);
    if(status&IMU_GYRO_DET)
      return true;
    else 
      return false;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::imu_get_accel_data(double *x,double *y,double *z)
{
  unsigned char data[6];

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_registers(DARWIN_IMU_ACCEL_X_L,data,6);

    *x=((double)((signed short int)(data[0]+(data[1]<<8))))/1000.0;
    *y=((double)((signed short int)(data[2]+(data[3]<<8))))/1000.0;
    *z=((double)((signed short int)(data[4]+(data[5]<<8))))/1000.0;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::imu_get_gyro_data(double *x,double *y,double *z)
{
  unsigned char data[6];

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_registers(DARWIN_IMU_GYRO_X_L,data,6);

    *x=((double)(data[0]+(data[1]<<8)))/1000.0;
    *y=((double)(data[2]+(data[3]<<8)))/1000.0;
    *z=((double)(data[4]+(data[5]<<8)))/1000.0;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

// motion manager interface
void CDarwinRobot::mm_set_period(double period_ms)
{
  unsigned short period;

  if(this->robot_device!=NULL)
  {
    period=(period_ms*1000.0);
    this->robot_device->write_word_register(DARWIN_MM_PERIOD_L,period);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

double CDarwinRobot::mm_get_period(void)
{
  unsigned short period;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_word_register(DARWIN_MM_PERIOD_L,&period);
    return ((double)period)/1000.0;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

unsigned char CDarwinRobot::mm_get_num_servos(void)
{
  if(this->robot_device!=NULL)
  {
    this->robot_device->read_byte_register(DARWIN_MM_NUM_SERVOS,&this->manager_num_servos);
    return this->manager_num_servos;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

unsigned int CDarwinRobot::mm_get_present_servos(void)
{
  unsigned int present_servos;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_registers(DARWIN_MM_PRESENT_SERVOS1,(unsigned char *)&present_servos,4);
    return present_servos;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::mm_start(void)
{
  if(this->robot_device!=NULL)
  {
    this->manager_status|=MANAGER_ENABLE;
    this->robot_device->write_byte_register(DARWIN_MM_CNTRL,this->manager_status);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::mm_stop(void)
{
  if(this->robot_device!=NULL)
  {
    this->manager_status&=(~MANAGER_ENABLE);
    this->robot_device->write_byte_register(DARWIN_MM_CNTRL,this->manager_status);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

bool CDarwinRobot::mm_is_running(void)
{
  unsigned char value;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_byte_register(DARWIN_MM_CNTRL,&value);
    if(value&MANAGER_ENABLE)
      return true;
    else
      return false;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::mm_enable_balance(void)
{
  if(this->robot_device!=NULL)
  {
    this->manager_status|=MANAGER_EN_BAL;
    this->robot_device->write_byte_register(DARWIN_MM_CNTRL,this->manager_status);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::mm_disable_balance(void)
{
  if(this->robot_device!=NULL)
  {
    this->manager_status&=(~MANAGER_EN_BAL);
    this->robot_device->write_byte_register(DARWIN_MM_CNTRL,this->manager_status);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

bool CDarwinRobot::mm_is_balance_enabled(void)
{
  unsigned char value;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_byte_register(DARWIN_MM_CNTRL,&value);
    if(value&MANAGER_EN_BAL)
      return true;
    else
      return false;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::mm_enable_power(void)
{
  if(this->robot_device!=NULL)
  {
    this->manager_status|=MANAGER_EN_PWR;
    this->robot_device->write_byte_register(DARWIN_MM_CNTRL,this->manager_status);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::mm_disable_power(void)
{
  if(this->robot_device!=NULL)
  {
    this->manager_status&=(~MANAGER_EN_PWR);
    this->robot_device->write_byte_register(DARWIN_MM_CNTRL,this->manager_status);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

bool CDarwinRobot::mm_is_power_enabled(void)
{
  unsigned char value;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_byte_register(DARWIN_MM_CNTRL,&value);
    if(value&MANAGER_EN_PWR)
      return true;
    else
      return false;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::mm_enable_servo(unsigned char servo_id)
{
  unsigned char value;

  if(this->robot_device!=NULL)
  {
    if(servo_id>MAX_NUM_SERVOS)
    {
      throw CDarwinRobotException(_HERE_,"Invalid servo identifier");
    }
    // get the current value
    this->robot_device->read_byte_register(DARWIN_MM_MODULE_EN0+servo_id/2,&value);
    if(servo_id%2)// odd servo
      value|=MANAGER_ODD_SER_EN;
    else// even servo
      value|=MANAGER_EVEN_SER_EN;
    this->robot_device->write_byte_register(DARWIN_MM_MODULE_EN0+servo_id/2,value);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::mm_disable_servo(unsigned char servo_id)
{
  unsigned char value;

  if(this->robot_device!=NULL)
  {
    if(servo_id>MAX_NUM_SERVOS)
    {
      throw CDarwinRobotException(_HERE_,"Invalid servo identifier");
    }
    // get the current value
    this->robot_device->read_byte_register(DARWIN_MM_MODULE_EN0+servo_id/2,&value);
    if(servo_id%2)// odd servo
      value&=(~MANAGER_ODD_SER_EN);
    else// even servo
      value&=(~MANAGER_EVEN_SER_EN);
    this->robot_device->write_byte_register(DARWIN_MM_MODULE_EN0+servo_id/2,value);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

bool CDarwinRobot::mm_is_servo_enabled(unsigned char servo_id)
{
  unsigned char value;

  if(this->robot_device!=NULL)
  {
    if(servo_id>MAX_NUM_SERVOS)
    {
      throw CDarwinRobotException(_HERE_,"Invalid servo identifier");
    }
    // get the current value
    this->robot_device->read_byte_register(DARWIN_MM_MODULE_EN0+servo_id/2,&value);
    if(servo_id%2)// odd servo
    {
      if(value&MANAGER_ODD_SER_EN)
        return true;
      else
        return false;
    }
    else// even servo
    {
      if(value&MANAGER_EVEN_SER_EN)
        return true;
      else
        return false;
    }
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::mm_assign_module(unsigned char servo_id, mm_mode_t mode)
{
  unsigned char value;

  if(this->robot_device!=NULL)
  {
    if(servo_id>MAX_NUM_SERVOS)
    {
      throw CDarwinRobotException(_HERE_,"Invalid servo identifier");
    }
    // get the current value
    this->robot_device->read_byte_register(DARWIN_MM_MODULE_EN0+servo_id/2,&value);
    if(servo_id%2)// odd servo
    {
      value&=(~MANAGER_ODD_SER_MOD);
      value|=(unsigned char)mode;
    }
    else// even servo
    {
      value&=(~MANAGER_EVEN_SER_MOD);
      value|=(((unsigned char)mode)<<4);
    }
    this->robot_device->write_byte_register(DARWIN_MM_MODULE_EN0+servo_id/2,value);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::mm_assign_module(std::string &servo,std::string &module)
{
  unsigned int i=0;

  for(i=0;i<MAX_NUM_SERVOS;i++)
  {
    if(servo==servo_names[i])
    {
      if(module=="action")
        this->mm_assign_module(i,DARWIN_MM_ACTION);
      else if(module=="walking")
        this->mm_assign_module(i,DARWIN_MM_WALKING);
      else if(module=="joints")
        this->mm_assign_module(i,DARWIN_MM_JOINTS);
      else if(module=="head")
        this->mm_assign_module(i,DARWIN_MM_HEAD);
      else if(module=="gripper")
        this->mm_assign_module(i,DARWIN_MM_GRIPPER);
      else
        this->mm_assign_module(i,DARWIN_MM_NONE);
      break;
    }
  }
}

mm_mode_t CDarwinRobot::mm_get_module(unsigned char servo_id)
{
  unsigned char value;

  if(this->robot_device!=NULL)
  {
    if(servo_id>MAX_NUM_SERVOS)
    {
      throw CDarwinRobotException(_HERE_,"Invalid servo identifier");
    }
    // get the current value
    this->robot_device->read_byte_register(DARWIN_MM_MODULE_EN0+servo_id/2,&value);
    if(servo_id%2)// odd servo
      return (mm_mode_t)(value&MANAGER_ODD_SER_MOD);
    else
      return (mm_mode_t)((value&MANAGER_EVEN_SER_MOD)>>4);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::mm_load_config(std::string &filename)
{
  darwin_config_t::servo_iterator iterator;
  unsigned int i=0;
  struct stat buffer;

  if(stat(filename.c_str(),&buffer)==0)
  {
    try{
      std::auto_ptr<darwin_config_t> cfg(darwin_config(filename.c_str(), xml_schema::flags::dont_validate));
      for(iterator=cfg->servo().begin(),i=0;iterator!=cfg->servo().end();iterator++,i++)
        this->mm_assign_module(iterator->name(),iterator->motion_module());
    }catch(const xml_schema::exception& e){
      std::ostringstream os;
      os << e;
      throw CDarwinRobotException(_HERE_,os.str());
    }
  }
  else
    throw CDarwinRobotException(_HERE_,"Configuration file does not exist");
}

std::vector<double> CDarwinRobot::mm_get_servo_angles(void)
{
  int i=0;
  short int values[MAX_NUM_SERVOS];
  std::vector<double> angles(MAX_NUM_SERVOS);

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_registers(DARWIN_MM_SERVO0_CUR_POS_L,(unsigned char *)values,MAX_NUM_SERVOS*2);

    for(i=0;i<MAX_NUM_SERVOS;i++)
      angles[i]=((double)values[i])/128.0;

    return angles;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

double CDarwinRobot::mm_get_servo_angle(unsigned char servo_id)
{
  unsigned short int value;
  double angle;

  if(this->robot_device!=NULL)
  {
    if(servo_id>MAX_NUM_SERVOS)
    {
      throw CDarwinRobotException(_HERE_,"Invalid servo identifier");
    }
    this->robot_device->read_word_register(DARWIN_MM_SERVO0_CUR_POS_L+servo_id/2,&value);
    angle=((double)value)/128.0;

    return angle;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::mm_set_servo_offset(unsigned char servo_id,double offset)
{
  if(offset<-9.0 || offset>9.0)
    throw CDarwinRobotException(_HERE_,"Desired offset out of range");
  else
  {
    if(servo_id>MAX_NUM_SERVOS)
      throw CDarwinRobotException(_HERE_,"Invalid servo identifier");
    else
    {
      this->robot_device->write_byte_register(DARWIN_MM_SERVO0_OFFSET+servo_id,(unsigned char)((int)(offset*16.0)));
      usleep(10000);
    }
  }
}

double CDarwinRobot::mm_get_servo_offset(unsigned char servo_id)
{
  unsigned char value;
  double offset;

  if(servo_id>MAX_NUM_SERVOS)
    throw CDarwinRobotException(_HERE_,"Invalid servo identifier");
  else
  {
    this->robot_device->read_byte_register(DARWIN_MM_SERVO0_OFFSET+servo_id,&value);
    offset=((double)((signed char)value))/16.0;

    return offset;
  }
}

std::vector<double> CDarwinRobot::mm_get_servo_offsets(void)
{
  int i=0;
  signed char values[MAX_NUM_SERVOS];
  std::vector<double> offsets(MAX_NUM_SERVOS);

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_registers(DARWIN_MM_SERVO0_OFFSET,(unsigned char *)values,MAX_NUM_SERVOS);

    for(i=0;i<MAX_NUM_SERVOS;i++)
      offsets[i]=((double)values[i])/16.0;

    return offsets;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::mm_set_balance_gains(double knee,double ankle_pitch,double hip_roll,double ankle_roll)
{

}

void CDarwinRobot::mm_get_balance_gains(double *knee,double *ankle_pitch,double *hip_roll,double *ankle_roll)
{

}

// motion action interface
void CDarwinRobot::action_load_page(unsigned char page_id)
{
  if(this->robot_device!=NULL)
    this->robot_device->write_byte_register(DARWIN_ACTION_PAGE,page_id);
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

unsigned char CDarwinRobot::action_get_current_page(void)
{
  unsigned char page_id;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_byte_register(DARWIN_ACTION_PAGE,&page_id);

    return page_id;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

unsigned char CDarwinRobot::action_get_current_step(void)
{
  unsigned char step_id;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_byte_register(DARWIN_ACTION_CNTRL,&step_id);

    return (step_id>>5);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::action_start(void)
{
  if(this->robot_device!=NULL)
  {
    this->action_status|=ACTION_START;
    this->robot_device->write_byte_register(DARWIN_ACTION_CNTRL,this->action_status);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::action_stop(void)
{
  if(this->robot_device!=NULL)
  {
    this->action_status|=ACTION_STOP;
    this->robot_device->write_byte_register(DARWIN_ACTION_CNTRL,this->action_status);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

bool CDarwinRobot::action_is_page_running(void)
{
  unsigned char status;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_byte_register(DARWIN_ACTION_CNTRL,&status);
    if(status&ACTION_STATUS)
      return true;
    else
      return false;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

/* walking interface */
void CDarwinRobot::walk_set_x_offset(double offset_m)
{
  unsigned char offset;

  if(this->robot_device!=NULL)
  {
    offset=(unsigned char)(offset_m*1000.0);
    this->robot_device->write_byte_register(DARWIN_WALK_X_OFFSET,offset);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

double CDarwinRobot::walk_get_x_offset(void)
{
  unsigned char offset;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_byte_register(DARWIN_WALK_X_OFFSET,&offset);
    return ((double)((signed char)offset))/1000.0;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::walk_set_y_offset(double offset_m)
{
  unsigned char offset;

  if(this->robot_device!=NULL)
  {
    offset=(unsigned char)(offset_m*1000.0);
    this->robot_device->write_byte_register(DARWIN_WALK_Y_OFFSET,offset);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

double CDarwinRobot::walk_get_y_offset(void)
{
  unsigned char offset;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_byte_register(DARWIN_WALK_Y_OFFSET,&offset);
    return ((double)((signed char)offset))/1000.0;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::walk_set_z_offset(double offset_m)
{
  unsigned char offset;

  if(this->robot_device!=NULL)
  {
    offset=(unsigned char)(offset_m*1000.0);
    this->robot_device->write_byte_register(DARWIN_WALK_Z_OFFSET,offset);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

double CDarwinRobot::walk_get_z_offset(void)
{
  unsigned char offset;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_byte_register(DARWIN_WALK_Z_OFFSET,&offset);
    return ((double)((signed char)offset))/1000.0;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::walk_set_roll_offset(double offset_rad)
{
  unsigned char offset;

  if(this->robot_device!=NULL)
  {
    offset=(unsigned char)((offset_rad*180.0/PI)*8.0);
    this->robot_device->write_byte_register(DARWIN_WALK_ROLL_OFFSET,offset);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

double CDarwinRobot::walk_get_roll_offset(void)
{
  unsigned char offset;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_byte_register(DARWIN_WALK_ROLL_OFFSET,&offset);
    return (((double)offset)*PI/180.0)/8.0;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::walk_set_pitch_offset(double offset_rad)
{
  unsigned char offset;

  if(this->robot_device!=NULL)
  {
    offset=(unsigned char)((offset_rad*180.0/PI)*8.0);
    this->robot_device->write_byte_register(DARWIN_WALK_PITCH_OFFSET,offset);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

double CDarwinRobot::walk_get_pitch_offset(void)
{
  unsigned char offset;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_byte_register(DARWIN_WALK_PITCH_OFFSET,&offset);
    return (((double)offset)*PI/180.0)/8.0;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::walk_set_yaw_offset(double offset_rad)
{
  unsigned char offset;

  if(this->robot_device!=NULL)
  {
    offset=(unsigned char)((offset_rad*180.0/PI)*8.0);
    this->robot_device->write_byte_register(DARWIN_WALK_YAW_OFFSET,offset);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

double CDarwinRobot::walk_get_yaw_offset(void)
{
  unsigned char offset;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_byte_register(DARWIN_WALK_YAW_OFFSET,&offset);
    return (((double)offset)*PI/180.0)/8.0;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::walk_set_hip_pitch_offset(double offset_rad)
{
  unsigned short int offset;

  if(this->robot_device!=NULL)
  {
    offset=(unsigned short int)((offset_rad*180.0/PI)*1024.0);
    this->robot_device->write_word_register(DARWIN_WALK_HIP_PITCH_OFF_L,offset);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

double CDarwinRobot::walk_get_hip_pitch_offset(void)
{
  unsigned short int offset;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_word_register(DARWIN_WALK_HIP_PITCH_OFF_L,&offset);
    return (((double)offset)*PI/180.0)/1024.0;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::walk_set_period(double period_s)
{
  unsigned short int period;

  if(this->robot_device!=NULL)
  {
    period=(unsigned short int)(period_s*1000.0);
    this->robot_device->write_word_register(DARWIN_WALK_PERIOD_TIME_L,period);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

double CDarwinRobot::walk_get_period(void)
{
  unsigned short int period;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_word_register(DARWIN_WALK_PERIOD_TIME_L,&period);
    return ((double)period)/1000.0;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::walk_set_ssp_dsp_ratio(double ratio)
{
  unsigned char ratio_int;

  if(this->robot_device!=NULL)
  {
    ratio_int=(unsigned char)(ratio*256.0);
    this->robot_device->write_byte_register(DARWIN_WALK_DSP_RATIO,ratio_int);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

double CDarwinRobot::walk_get_ssp_dsp_ratio(void)
{
  unsigned char ratio_int;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_byte_register(DARWIN_WALK_DSP_RATIO,&ratio_int);
    return ((double)ratio_int)/256.0;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::walk_set_fwd_bwd_ratio(double ratio)
{
  unsigned char ratio_int;

  if(this->robot_device!=NULL)
  {
    ratio_int=(unsigned char)(ratio*256.0);
    this->robot_device->write_byte_register(DARWIN_WALK_STEP_FW_BW_RATIO,ratio_int);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

double CDarwinRobot::walk_get_fwd_bwd_ratio(void)
{
  unsigned char ratio_int;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_byte_register(DARWIN_WALK_STEP_FW_BW_RATIO,&ratio_int);
    return ((double)ratio_int)/256.0;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::walk_set_foot_height(double height_m)
{
  unsigned char height;

  if(this->robot_device!=NULL)
  {
    height=(unsigned char)(height_m*1000.0);
    this->robot_device->write_byte_register(DARWIN_WALK_FOOT_HEIGHT,height);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

double CDarwinRobot::walk_get_foot_height(void)
{
  unsigned char height;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_byte_register(DARWIN_WALK_FOOT_HEIGHT,&height);
    return ((double)height)/1000.0;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::walk_set_left_right_swing(double swing_m)
{
  unsigned char swing;

  if(this->robot_device!=NULL)
  {
    swing=(unsigned char)(swing_m*1000.0);
    this->robot_device->write_byte_register(DARWIN_WALK_SWING_RIGHT_LEFT,swing);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

double CDarwinRobot::walk_get_left_right_swing(void)
{
  unsigned char swing;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_byte_register(DARWIN_WALK_SWING_RIGHT_LEFT,&swing);
    return ((double)swing)/1000.0;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::walk_set_top_down_swing(double swing_m)
{
  unsigned char swing;

  if(this->robot_device!=NULL)
  {
    swing=(unsigned char)(swing_m*1000.0);
    this->robot_device->write_byte_register(DARWIN_WALK_SWING_TOP_DOWN,swing);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

double CDarwinRobot::walk_get_top_down_swing(void)
{
  unsigned char swing;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_byte_register(DARWIN_WALK_SWING_TOP_DOWN,&swing);
    return ((double)swing)/1000.0;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::walk_set_pelvis_offset(double offset_rad)
{
  unsigned char offset;

  if(this->robot_device!=NULL)
  {
    offset=(unsigned char)((offset_rad*180.0/PI)*8.0);
    this->robot_device->write_byte_register(DARWIN_WALK_PELVIS_OFFSET,offset);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

double CDarwinRobot::walk_get_pelvis_offset(void)
{
  unsigned char offset;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_byte_register(DARWIN_WALK_PELVIS_OFFSET,&offset);
    return (((double)offset)*PI/180.0)/8.0;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::walk_set_arm_swing_gain(double gain)
{
  unsigned char gain_int;

  if(this->robot_device!=NULL)
  {
    gain_int=(unsigned char)(gain*32.0);
    this->robot_device->write_byte_register(DARWIN_WALK_ARM_SWING_GAIN,gain_int);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

double CDarwinRobot::walk_get_arm_swing_gain(void)
{
  unsigned char gain_int;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_byte_register(DARWIN_WALK_ARM_SWING_GAIN,&gain_int);
    return ((double)gain_int)/32.0;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::walk_set_trans_speed(double speed_m_s)
{
  unsigned char speed;

  if(this->robot_device!=NULL)
  {
    speed=(unsigned char)(speed_m_s*1000.0);
    this->robot_device->write_byte_register(DARWIN_WALK_MAX_VEL,speed);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

double CDarwinRobot::walk_get_trans_speed(void)
{
  unsigned char speed;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_byte_register(DARWIN_WALK_MAX_VEL,&speed);
    return ((double)speed)/1000.0;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::walk_set_rot_speed(double speed_rad_s)
{
  unsigned char speed;

  if(this->robot_device!=NULL)
  {
    speed=(unsigned char)((speed_rad_s*180.0/PI)*8.0);
    this->robot_device->write_byte_register(DARWIN_WALK_MAX_ROT_VEL,speed);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

double CDarwinRobot::walk_get_rot_speed(void)
{
  unsigned char speed;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_byte_register(DARWIN_WALK_MAX_ROT_VEL,&speed);
    return (((double)speed)*PI/180.0)/8.0;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::walk_start(void)
{
  if(this->robot_device!=NULL)
    this->robot_device->write_byte_register(DARWIN_WALK_CNTRL,WALK_START);
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::walk_stop(void)
{
  if(this->robot_device!=NULL)
    this->robot_device->write_byte_register(DARWIN_WALK_CNTRL,WALK_STOP);
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

bool CDarwinRobot::is_walking(void)
{
  unsigned char value;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_byte_register(DARWIN_WALK_CNTRL,&value);
    if(value&WALK_STATUS)
      return true;
    else
      return false;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::walk_set_x_step(double step_m)
{
  unsigned char step;

  if(this->robot_device!=NULL)
  {
    step=(unsigned char)(step_m*1000.0);
    this->robot_device->write_byte_register(DARWIN_WALK_STEP_FW_BW,step);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

double CDarwinRobot::walk_get_x_step(void)
{
  unsigned char step;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_byte_register(DARWIN_WALK_STEP_FW_BW,&step);
    return ((double)step)/1000.0;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::walk_set_y_step(double step_m)
{
  unsigned char step;

  if(this->robot_device!=NULL)
  {
    step=(unsigned char)(step_m*1000.0);
    this->robot_device->write_byte_register(DARWIN_WALK_STEP_LEFT_RIGHT,step);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

double CDarwinRobot::walk_get_y_step(void)
{
  unsigned char step;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_byte_register(DARWIN_WALK_STEP_LEFT_RIGHT,&step);
    return ((double)step)/1000.0;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::walk_set_turn_step(double step_rad)
{
  unsigned char step;

  if(this->robot_device!=NULL)
  {
    step=(unsigned char)((step_rad*180.0/PI)*8.0);
    this->robot_device->write_byte_register(DARWIN_WALK_STEP_DIRECTION,step);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

double CDarwinRobot::walk_get_turn_step(void)
{
  unsigned char step;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_byte_register(DARWIN_WALK_STEP_DIRECTION,&step);
    return (((double)step)*PI/180.0)/8.0;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

// joint motion interface
void CDarwinRobot::joints_load(joints_grp_t group,std::vector<unsigned char> &servos,std::vector<double> &angles,std::vector<double> &speeds,std::vector<double> &accels)
{
  unsigned int group_servos=0,i;
  unsigned short int angle;
  unsigned char speed,accel;

  if(this->robot_device!=NULL)
  {
    if(servos.size()!=angles.size() || servos.size()!=speeds.size() || servos.size()!=accels.size())
      throw CDarwinRobotException(_HERE_,"Invalid input vector sizes");
    for(i=0;i<servos.size();i++)
    {
      group_servos|=(0x00000001<<servos[i]);
      angle=(unsigned short int)(angles[i]*128.0);
      this->robot_device->write_word_register(DARWIN_JOINT_SERVO0_ANGLE_L+servos[i]*2,angle);
      speed=(unsigned char)(speeds[i]);
      this->robot_device->write_byte_register(DARWIN_JOINT_SERVO0_SPEED+servos[i],speed);
      accel=(unsigned char)(accels[i]);
      this->robot_device->write_byte_register(DARWIN_JOINT_SERVO0_ACCEL+servos[i],accel);
    }
    this->robot_device->write_registers(DARWIN_JOINT_GRP0_SERVOS0+((unsigned short int)group)*5,(unsigned char *)&group_servos,4);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

std::vector<unsigned char> CDarwinRobot::joints_get_group_servos(joints_grp_t group)
{

}

std::vector<double> CDarwinRobot::joints_get_group_angles(joints_grp_t group)
{

}

std::vector<double> CDarwinRobot::joints_get_group_speeds(joints_grp_t group)
{

}

std::vector<double> CDarwinRobot::joints_get_group_accels(joints_grp_t group)
{

}


void CDarwinRobot::joints_start(joints_grp_t group)
{
  if(this->robot_device!=NULL)
    this->robot_device->write_byte_register(DARWIN_JOINT_GRP0_CNTRL+((unsigned short int)group*5),JOINT_START);
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::joints_stop(joints_grp_t group)
{
  if(this->robot_device!=NULL)
    this->robot_device->write_byte_register(DARWIN_JOINT_GRP0_CNTRL+((unsigned short int)group*5),JOINT_STOP);
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

bool CDarwinRobot::joints_are_moving(joints_grp_t group)
{
  unsigned char value;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_byte_register(DARWIN_JOINT_GRP0_CNTRL+((unsigned short int)group*5),&value);
    if(value&JOINT_STATUS)
      return true;
    else
      return false;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

// head tracking interface
void CDarwinRobot::head_set_pan_pid(double p,double i,double d,double i_clamp)
{
  unsigned short int value;

  if(this->robot_device!=NULL)
  {
    value=p*65536.0;
    this->robot_device->write_word_register(DARWIN_HEAD_PAN_P_L,value);
    usleep(100000);
    value=i*65536.0;
    this->robot_device->write_word_register(DARWIN_HEAD_PAN_I_L,value);
    usleep(100000);
    value=d*65536.0;
    this->robot_device->write_word_register(DARWIN_HEAD_PAN_D_L,value);
    usleep(100000);
    value=i_clamp*128.0;
    this->robot_device->write_word_register(DARWIN_HEAD_PAN_I_CLAMP_L,value);
    usleep(100000);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::head_get_pan_pid(double *p,double *i,double *d,double *i_clamp)
{
  unsigned short int value;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_word_register(DARWIN_HEAD_PAN_P_L,&value);
    *p=((double)value)/65536.0;
    this->robot_device->read_word_register(DARWIN_HEAD_PAN_I_L,&value);
    *i=((double)value)/65536.0;
    this->robot_device->read_word_register(DARWIN_HEAD_PAN_D_L,&value);
    *d=((double)value)/65536.0;
    this->robot_device->read_word_register(DARWIN_HEAD_PAN_I_CLAMP_L,&value);
    *i_clamp=((double)value)/128.0;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::head_set_tilt_pid(double p,double i,double d,double i_clamp)
{
  unsigned short int value;

  if(this->robot_device!=NULL)
  {
    value=p*65536.0;
    this->robot_device->write_word_register(DARWIN_HEAD_TILT_P_L,value);
    usleep(100000);
    value=i*65536.0;
    this->robot_device->write_word_register(DARWIN_HEAD_TILT_I_L,value);
    usleep(100000);
    value=d*65536.0;
    this->robot_device->write_word_register(DARWIN_HEAD_TILT_D_L,value);
    usleep(100000);
    value=i_clamp*128.0;
    this->robot_device->write_word_register(DARWIN_HEAD_TILT_I_CLAMP_L,value);
    usleep(100000);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::head_get_tilt_pid(double *p,double *i,double *d,double *i_clamp)
{
  unsigned short int value;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_word_register(DARWIN_HEAD_TILT_P_L,&value);
    *p=((double)value)/65536.0;
    this->robot_device->read_word_register(DARWIN_HEAD_TILT_I_L,&value);
    *i=((double)value)/65536.0;
    this->robot_device->read_word_register(DARWIN_HEAD_TILT_D_L,&value);
    *d=((double)value)/65536.0;
    this->robot_device->read_word_register(DARWIN_HEAD_TILT_I_CLAMP_L,&value);
    *i_clamp=((double)value)/128.0;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::head_set_pan_range(double max,double min)
{
  unsigned short int value;

  if(this->robot_device!=NULL)
  {
    value=max*128.0;   
    this->robot_device->write_word_register(DARWIN_HEAD_MAX_PAN_L,value);
    value=min*128.0;   
    this->robot_device->write_word_register(DARWIN_HEAD_MIN_PAN_L,value);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::head_get_pan_range(double *max,double *min)
{
  unsigned short int value;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_word_register(DARWIN_HEAD_MAX_PAN_L,&value);
    *max=((double)value)/128.0; 
    this->robot_device->read_word_register(DARWIN_HEAD_MIN_PAN_L,&value);
    *min=((double)value)/128.0; 
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::head_set_tilt_range(double max,double min)
{
  unsigned short int value;

  if(this->robot_device!=NULL)
  {
    value=max*128.0;  
    this->robot_device->write_word_register(DARWIN_HEAD_MAX_TILT_L,value);
    value=min*128.0;
    this->robot_device->write_word_register(DARWIN_HEAD_MIN_TILT_L,value);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::head_get_tilt_range(double *max,double *min)
{
  unsigned short int value;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_word_register(DARWIN_HEAD_MAX_TILT_L,&value);
    *max=((double)value)/128.0;
    this->robot_device->read_word_register(DARWIN_HEAD_MIN_TILT_L,&value);
    *min=((double)value)/128.0;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::head_start_tracking(double pan,double tilt)
{
  unsigned short int value;

  if(this->robot_device!=NULL)
  {
    value=pan*128.0;
    this->robot_device->write_word_register(DARWIN_HEAD_PAN_TARGET_L,value);
    value=tilt*128.0;
    this->robot_device->write_word_register(DARWIN_HEAD_TILT_TARGET_L,value);
    this->robot_device->write_byte_register(DARWIN_HEAD_CNTRL,HEAD_START);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::head_stop_tracking(void)
{
  if(this->robot_device!=NULL)
    this->robot_device->write_byte_register(DARWIN_HEAD_CNTRL,HEAD_STOP);
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

bool CDarwinRobot::head_is_tracking(void)
{ 
  unsigned char value;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_byte_register(DARWIN_HEAD_CNTRL,&value);
    if(value&HEAD_STATUS)
      return true;
    else
      return false;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::head_set_new_target(double pan,double tilt)
{
  unsigned short int value;

  if(this->robot_device!=NULL)
  {
    value=pan*128.0;
    this->robot_device->write_word_register(DARWIN_HEAD_PAN_TARGET_L,value);
    value=tilt*128.0;
    this->robot_device->write_word_register(DARWIN_HEAD_TILT_TARGET_L,value);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::head_get_current_target(double *pan,double *tilt)
{
  unsigned short int value;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_word_register(DARWIN_HEAD_PAN_TARGET_L,&value);
    *pan=((double)value)/128.0;
    this->robot_device->read_word_register(DARWIN_HEAD_TILT_TARGET_L,&value);
    *tilt=((double)value)/128.0;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

//Smart charger interface
void CDarwinRobot::smart_charger_enable(void)
{
  if(this->robot_device!=NULL)
  {
    this->smart_charger_status|=SMART_CHARGER_EN;
    this->robot_device->write_byte_register(DARWIN_SMART_CHARGER_CNTRL,this->smart_charger_status);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

/* Disables the smart charger module
 */
void CDarwinRobot::smart_charger_disable(void)
{
  if(this->robot_device!=NULL)
  {
    this->smart_charger_status&=(~SMART_CHARGER_EN);
    this->robot_device->write_byte_register(DARWIN_SMART_CHARGER_CNTRL,this->smart_charger_status);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

bool CDarwinRobot::is_smart_charger_detected(void)
{
  unsigned char status;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_byte_register(DARWIN_SMART_CHARGER_CNTRL,&status);
    if(status&SMART_CHARGER_DET)
      return true;
    else 
      return false;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}
bool CDarwinRobot::is_smart_charger_det_and_en(void)
{
  unsigned char status;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_byte_register(DARWIN_SMART_CHARGER_CNTRL,&status);
    if(status&SMART_CHARGER_EN && status&SMART_CHARGER_DET)
      return true;
    else 
      return false;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}


void CDarwinRobot::smart_charger_set_period(double period)
{
  unsigned short int period_ms;

  if(this->robot_device!=NULL)
  {
    period_ms=period*1000;
    if(period_ms>=1400 && period_ms<=1600)
      this->robot_device->write_word_register(DARWIN_SMART_CHARGER_PERIOD_L,period_ms);
    else
      std::cout <<" Invalid period value" << std::endl;
      //throw CDarwinRobotException(_HERE_,"Invalid period value");
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}


double CDarwinRobot::smart_charger_get_period(void)
{
  unsigned short int period_ms;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_word_register(DARWIN_SMART_CHARGER_PERIOD_L,&period_ms);
    
    return ((double)period_ms)/1000;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

/*
void CDarwinRobot::smart_charger_get_period(double *period)
{
  unsigned short int period_ms;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_word_register(DARWIN_SMART_CHARGER_PERIOD_L,&period_ms);
    *period= ((double)period_ms)/1000;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}
*/
TChargerData CDarwinRobot::smart_charger_get_data(void)
{
 // unsigned short charger_data[3];
  TChargerData smart_charger_data;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_registers(DARWIN_SMART_CHARGER_AVG_TIME_EMPTY_L,(uint8_t *)&smart_charger_data,6);
    return smart_charger_data;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}


unsigned int CDarwinRobot::smart_charger_read_time_empty(void)
{
  unsigned char data[2];
  unsigned int avg_time_empty;
  
    if(this->robot_device!=NULL)
  {
    this->robot_device->read_byte_register(DARWIN_SMART_CHARGER_AVG_TIME_EMPTY_L,data);
    avg_time_empty=(data[0]+(data[1]<<8));
    return avg_time_empty;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

unsigned char CDarwinRobot::smart_charger_read_control(void)
{
  unsigned char smart_charger_control;
    if(this->robot_device!=NULL)
  {
    this->robot_device->read_byte_register(DARWIN_SMART_CHARGER_CNTRL,&smart_charger_control);
    return smart_charger_control;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

unsigned char CDarwinRobot::smart_charger_read_status(void)
{
  unsigned char smart_charger_stat;
    if(this->robot_device!=NULL)
  {
    this->robot_device->read_byte_register(DARWIN_SMART_CHARGER_STATUS,&smart_charger_stat);
    return smart_charger_stat;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

//Set limit input current
void CDarwinRobot::smart_charger_range_current(double min_current, double max_current)
{
  this->MAX_limit_current=max_current; 
  this->MIN_limit_current=min_current;
}
void CDarwinRobot::smart_charger_set_limit_current(double limit_current)
{
  unsigned short int limit_current_ma;

  if(this->robot_device!=NULL)
  {
    //if(limit_current > MIN_limit_current && limit_current < MAX_limit_current){
    if(limit_current > 0.255 && limit_current < 1.0){
      limit_current_ma=limit_current*1000;
      this->robot_device->write_word_register(DARWIN_SMART_CHARGER_LIMIT_CURRENT_L,limit_current_ma);
    }else{
      std::cout <<" Invalid current value" << std::endl;
    }
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}


CDarwinRobot::~CDarwinRobot()
{
  if(this->robot_device!=NULL)
  {
    this->dyn_server->free_device(this->robot_id);
  }
}

